%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% Algorithm.m
%
classdef (Abstract) Algorithm < handle

  properties (GetAccess = protected, SetAccess = private)
    mcdcStrategy
    kernelGeneratorStrategyHandle
    stateModel
    obsModel
  end

  properties (Access = private)
    likelihoodCalculator
    stateNoise
    obsNoise
    numDraw

    iterationStrategy

  end

  methods (Access = protected)

    function this = Algorithm(kernelGeneratorStrategy, mcdcStrategy)
      this.mcdcStrategy = mcdcStrategy(this);
      this.kernelGeneratorStrategyHandle = kernelGeneratorStrategy;
      this.numDraw = 1;

      this.iterationStrategy = IterationStrategyProbabilistic(0.25);
    end

  end

  methods (Access = public)

    function SetPMCMCProbability(this, probability)
      if probability < 1
        this.iterationStrategy = IterationStrategyProbabilistic(probability);
      else
        this.iterationStrategy = IterationStrategyDefault();
      end
    end

    function strategy = GetStrategy(this)
      strategy = this.mcdcStrategy;
    end

    function strategy = GetKernelGeneratorStrategyHandle(this)
      strategy = this.kernelGeneratorStrategyHandle;
    end

    function SetStateModel(this, stateModel)
      this.stateModel = stateModel;
    end

    function SetObsModel(this, obsModel)
      this.obsModel = obsModel;
    end

    function SetLikelihoodCalculator(this, likelihoodCalculator)
      this.likelihoodCalculator = likelihoodCalculator;
    end

    function SetStateNoise(this, stateNoise)
      this.stateNoise = stateNoise;
    end

    function SetObsNoise(this, obsNoise)
      this.obsNoise = obsNoise;
    end

    function SetNumDraw(this, numDraw)
      this.numDraw = numDraw;
    end

    function Execute(this, in, out, aspect)
      xdim = in.XDim();
      ydim = in.YDim();
      T = in.T();

      this.stateModel.Initialize(in, out, aspect);
      this.obsModel.Initialize(in, out, aspect);

      gridData = GridData(in.xGrids);
      this.mcdcStrategy.SetGridData(gridData);

      % Start
      aspect.Start(this, in, out);
      cleanupObj = onCleanup(@() aspect.End());

      % Grid Check
      CheckGridTransformation(VectorValuedFunction(in.stateMeanFuncs), gridData, aspect);

      %%%
      %%% MCMC Iteration
      %%%

      this.iterationStrategy.SetUp( ...
        this, ...
        this.mcdcStrategy, ...
        this.stateModel, ...
        this.obsModel, ...
        this.likelihoodCalculator, ...
        this.stateNoise, ...
        this.obsNoise, ...
        this.numDraw ...
      );

      for j = (out.j + 1):in.J

        this.iterationStrategy.ProcessIteration(j, in, out, aspect, gridData);

      end

    end

    function result = DrawGPSurface(this, gridData, pivot, means, gramMatrices)
      dim = size(means, 1);
      result = zeros([dim gridData.CountPerDim]);
      indices = this.getIndicesFromGridData(gridData, pivot);
      for d = 1:dim
        for i = 1:size(indices, 1)
          result(d, indices(i, :)) = mvnrnd(means(d, indices(i, :)), gramMatrices{d});
        end
      end
    end

    function result = CalcLogProbability(this, values, gridData, pivot, means, gramMatrices)
      threshold = 1e-10;
      dim = size(means, 1);
      result = 0;
      indices = this.getIndicesFromGridData(gridData, pivot);
      for d = 1:dim
        sigma = gramMatrices{d};
        eigmin = min(eig(sigma));
        if eigmin < threshold
          sigma = sigma + (threshold - eigmin) .* eye(size(sigma));
        end
        for i = 1:size(indices, 1)
          x = values(d, indices(i, :));
          mu = means(d, indices(i, :));
          result = result + logmvnpdf(x, mu, sigma);
        end
      end
    end

    function gramMatrixGrid = getGramMatrixGrid(this, in, gridData, aspect)
      aspect.Log('  Creating GramMatrix using all dims...');
      gramMatrixGrid = gridData.GetFlatMatrix();
    end

  end

  methods (Access = protected)

    function indices = getIndicesFromGridData(this, gridData, pivot)
      indices = gridData.GetAllIndices();
    end

  end

end
