%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% MCDCStrategy.m
%

classdef (Abstract) MCDCStrategy < handle

  properties (Access = protected)
    algorithm
    prevGramMatState
    prevGramMatObs
    gridData
    initialStateMean
    initialObsMean
  end

  methods (Access = protected)

    function this = MCDCStrategy(algorithm)
      this.algorithm = algorithm;
    end

  end

  methods (Abstract, Access = public)

    StateMean(this, j, in, out)

    ObsMean(this, j, in, out)

    StateSigma(this, j, gramMatrix)

    ObsSigma(this, j, gramMatrix)

    AcceptanceLogProbability(this, j, in, out, logLikelihood, stateModel, obsModel)

  end

  methods (Access = public)

    function result = InitialStateMean(this)
      result = this.initialStateMean;
    end

    function result = InitialObsMean(this)
      result = this.initialObsMean;
    end

    function SetGridData(this, gridData)
      this.gridData = gridData;
    end

    function SaveGramMatrices(this, gramMatState, gramMatObs)
      this.prevGramMatState = gramMatState;
      this.prevGramMatObs   = gramMatObs;
    end

  end

  methods (Access = protected)

    function SetInitialStateMean(this, in)
      gridPoints = this.gridData.GetFlatMatrix();
      dims = [size(in.stateMeanFuncs, 2) this.gridData.CountPerDim];
      this.initialStateMean = zeros(dims);
      for d = 1:dims(1)
        this.initialStateMean(d, :) = in.stateMeanFuncs{d}(gridPoints);
      end
    end

    function SetInitialObsMean(this, in)
      gridPoints = this.gridData.GetFlatMatrix();
      dims = [size(in.obsMeanFuncs, 2) this.gridData.CountPerDim];
      this.initialObsMean = zeros(dims);
      for d = 1:dims(1)
        this.initialObsMean(d, :) = in.obsMeanFuncs{d}(gridPoints);
      end
    end

    function result = LikelihoodContribution(this, j, out, logLikelihood)
      result = logLikelihood - out.loglikAt(j - 1);
    end

    function result = PriorContribution(this, j, in, out, stateModel, obsModel)
      result = 0;
if isa(stateModel, 'GaussianProcessModel')
      result = result + this.PriorContributionImpl( ...
          j, in, out, stateModel, reshape(out.FVAt(j - 1), size(stateModel.eqnVals)), this.initialStateMean, this.prevGramMatState);
end
if isa(obsModel, 'GaussianProcessModel')
      result = result + this.PriorContributionImpl( ...
          j, in, out, obsModel, reshape(out.GVAt(j - 1), size(obsModel.eqnVals)), this.initialObsMean, this.prevGramMatObs);
end
    end

    function result = TransitionContribution(this, j, in, out, stateModel, obsModel)
      result = 0;
if isa(stateModel, 'GaussianProcessModel')
      result = result + this.TransitionContributionImpl( ...
          j, in, out, stateModel, reshape(out.FVAt(j - 1), size(stateModel.eqnVals)), this.prevGramMatState);
end
if isa(obsModel, 'GaussianProcessModel')
      result = result + this.TransitionContributionImpl( ...
          j, in, out, obsModel, reshape(out.GVAt(j - 1), size(obsModel.eqnVals)), this.prevGramMatObs);
end
    end

  end

  methods (Access = private)

    function result = PriorContributionImpl(this, j, in, out, model, prevEqnVals, initialMean, prevGramMatrix)
      result = this.algorithm.CalcLogProbability( ...
          model.eqnVals, this.gridData, in.gridDimForGramMatrix, initialMean, model.gramMatrix);
      result = result - this.algorithm.CalcLogProbability( ...
          prevEqnVals, this.gridData, in.gridDimForGramMatrix, initialMean, prevGramMatrix);
    end

    function result = TransitionContributionImpl(this, j, in, out, model, prevEqnVals, prevGramMatrix)
      result = this.algorithm.CalcLogProbability( ...
          prevEqnVals, this.gridData, in.gridDimForGramMatrix, model.eqnVals, prevGramMatrix);
      result = result - this.algorithm.CalcLogProbability( ...
          model.eqnVals, this.gridData, in.gridDimForGramMatrix, prevEqnVals, model.gramMatrix);
    end

  end

end
