  %%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% MultipleKnownSequences.m
%
classdef MultipleKnownSequences < LikelihoodCalculator

  properties (Access = private)
    x
    xaux
    u
    y
    q
    r
    K
    currentLogLikelihood
    outIndex
  end

  methods (Access = public)

    function this = MultipleKnownSequences(x, xaux, u, y, q, r, K)
      this.x    = x;
      this.xaux = xaux;
      this.u    = u;
      this.y    = y;
      this.q    = q;
      this.r    = r;
      this.K    = K;

      this.outIndex = 1;
    end

    function logLikelihood = LogLikelihood( ...
        this, ...
        fnState, ...
        fnObs, ...
        stateTransitionNoise, ...
        observationNoise)

      % Calculate the joint probability of given sequences.

      logLikelihood = 0;
      for i = 1:length(this.x)
        if nargin < 4
          logLikelihood = logLikelihood + ...
              this.LogLikelihoodOfSequence( ...
                  i, ...
                  fnState, ...
                  fnObs);
        else
          logLikelihood = logLikelihood + ...
              this.LogLikelihoodOfSequence( ...
                  i, ...
                  fnState, ...
                  fnObs, ...
                  stateTransitionNoise, ...
                  observationNoise);
        end
      end

      % We record logLikelihood to the property of this object.
      % It's just for logging purpose (see OnAccept below.)
      this.currentLogLikelihood = logLikelihood;
    end

    function OnAccept(this, j, in, out, aspect)
      out.AddXE(this.x{this.outIndex});
      out.AddYE(this.y{this.outIndex});
      out.AddLoglik(this.currentLogLikelihood);
    end

    function logLikelihood = MHStateLogLikelihood( ...
        this, ...
        fnState, ...
        stateTransitionNoise)

      % Calculate the joint probability of given sequences.

      logLikelihood = 0;
      for i = 1:length(this.x)
        if nargin < 3
          logLikelihood = logLikelihood + ...
              this.LogLikelihoodOfStateSequence( ...
                  i, ...
                  fnState);
        else
          logLikelihood = logLikelihood + ...
              this.LogLikelihoodOfStateSequence( ...
                  i, ...
                  fnState, ...
                  stateTransitionNoise);
        end
      end
    end

    function logLikelihood = MHObsLogLikelihood( ...
        this, ...
        fnObs, ...
        observationNoise)

      % Calculate the joint probability of given sequences.

      logLikelihood = 0;
      for i = 1:length(this.x)
        if nargin < 3
          logLikelihood = logLikelihood + ...
              this.LogLikelihoodOfObsSequence( ...
                  i, ...
                  fnObs);
        else
          logLikelihood = logLikelihood + ...
              this.LogLikelihoodOfObsSequence( ...
                  i, ...
                  fnObs, ...
                  observationNoise);
        end
      end
    end
  end

  methods (Access = private)

    function logLikelihood = LogLikelihoodOfSequence( ...
        this, ...
        index, ...
        fnState, ...
        fnObs, ...
        stateTransitionNoise, ...
        observationNoise)

      % Calculate the joint probability of the sequence x(1:T) and y(1:T)
      % under the state space model:
      %   x(t) = fnState([x(t-1) xaux(t-1)]) + u(t) + q
      %   y(t+K) = fnObs(x(t)) + r
      %   q ~ stateTransitionNoise
      %   r ~ observationNoise
      % Notes
      %  1. for q and r, normal distribution is used unless otherwise specified.
      %  2. K is introduced as 'observation offset,' but we use K=0 usually.

      x = this.x{index};
      y = this.y{index};
      q = this.q;
      r = this.r;
      K = this.K;

      if nargin < 5
        stateTransitionNoise = NormalDistribution(0, q, size(x, 2));
        observationNoise = NormalDistribution(0, r, size(y, 2));
      end

      if ~isempty(this.xaux)
        xIn = [x this.xaux{index}];
      else
        xIn = x;
      end
      xRes = x(2:end, :) - fnState(xIn(1:end-1, :));
      if ~isempty(this.u)
        u = this.u{index};
        xRes = xRes - u(2:end, :);
      end
      xLogProb = stateTransitionNoise.LogPdf(xRes);

      yRes = y(K+1:end, :) - fnObs(xIn(1:end-K, :));
      yLogProb = observationNoise.LogPdf(yRes);

      logLikelihood = sum(xLogProb) + sum(yLogProb);

    end

    function logLikelihood = LogLikelihoodOfStateSequence( ...
        this, ...
        index, ...
        fnState, ...
        stateTransitionNoise)

      % Calculate the joint probability of the sequence x(1:T) and y(1:T)
      % under the state space model:
      %   x(t) = fnState([x(t-1) xaux(t-1)]) + u(t) + q
      %   y(t+K) = fnObs(x(t)) + r
      %   q ~ stateTransitionNoise
      %   r ~ observationNoise
      % Notes
      %  1. for q and r, normal distribution is used unless otherwise specified.
      %  2. K is introduced as 'observation offset,' but we use K=0 usually.

      x = this.x{index};
      q = this.q;
      K = this.K;

      if nargin < 4
        stateTransitionNoise = NormalDistribution(0, q, size(x, 2));
      end

      if ~isempty(this.xaux)
        xIn = [x this.xaux{index}];
      else
        xIn = x;
      end
      xRes = x(2:end, :) - fnState(xIn(1:end-1, :));
      if ~isempty(this.u)
        u = this.u{index};
        xRes = xRes - u(2:end, :);
      end
      xLogProb = stateTransitionNoise.LogPdf(xRes);

      logLikelihood = sum(xLogProb);

    end

    function logLikelihood = LogLikelihoodOfObsSequence( ...
        this, ...
        index, ...
        fnObs, ...
        observationNoise)

      % Calculate the joint probability of the sequence x(1:T) and y(1:T)
      % under the state space model:
      %   x(t) = fnState([x(t-1) xaux(t-1)]) + u(t) + q
      %   y(t+K) = fnObs(x(t)) + r
      %   q ~ stateTransitionNoise
      %   r ~ observationNoise
      % Notes
      %  1. for q and r, normal distribution is used unless otherwise specified.
      %  2. K is introduced as 'observation offset,' but we use K=0 usually.

      x = this.x{index};
      y = this.y{index};
      r = this.r;
      K = this.K;

      if nargin < 4
        observationNoise = NormalDistribution(0, r, size(y, 2));
      end

      if ~isempty(this.xaux)
        xIn = [x this.xaux{index}];
      else
        xIn = x;
      end

      yRes = y(K+1:end, :) - fnObs(xIn(1:end-K, :));
      yLogProb = observationNoise.LogPdf(yRes);

      logLikelihood = sum(yLogProb);

    end

  end

end
