%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% PMCMCParticleFilter.m
%

classdef PMCMCParticleFilter < ParticleFilter

  properties (Access = private)
    ji
  end

  methods (Access = public)

    function this = PMCMCParticleFilter(x0, xaux, u, y, q, r, N, K)
      this = this@ParticleFilter(x0, xaux, u, y, q, r, N, K);
    end

    function Estimate( ...
        this, ...
        f, ...
        h, ...
        stateTransitionNoise, ...
        observationNoise)
      Estimate@ParticleFilter( ...
          this, ...
          f, ...
          h, ...
          stateTransitionNoise, ...
          observationNoise);
      T = size(this.y, 1);
      % Generate the trajectories from ancestor indices
      ind = this.ancestorIndicies(:, T);
      for(t = T-1:-1:1)
          this.Particles(:, t) = this.Particles(ind, t);
          ind = this.ancestorIndicies(ind, t);
      end
      % Sample from trajectories
      this.ji = find(rand(1) < cumsum(this.Weights(:, T - this.K)), 1, 'first');
   % 161            xe = reshape(pf.Particles(ji, :, :), size(pf.Particles, 2), size(pf.Particles, 3));

    end

    function logLikelihood = MHStateLogLikelihood( ...
        this, ...
        fnState, ...
        stateTransitionNoise)

      % Calculate the joint probability of the sequence x(1:T) and y(1:T)
      % under the state space model:
      %   x(t) = fnState([x(t-1) xaux(t-1)]) + u(t) + q
      %   y(t+K) = fnObs(x(t)) + r
      %   q ~ stateTransitionNoise
      %   r ~ observationNoise
      % Notes
      %  1. for q and r, normal distribution is used unless otherwise specified.
      %  2. K is introduced as 'observation offset,' but we use K=0 usually.

      xe = this.GetCurrentXE();

      if nargin < 3
        stateTransitionNoise = NormalDistribution(0, this.q, size(xe, 2));
      end
      xIn      = [xe this.xaux];
      xRes     = xe(2:end, :) - fnState(xIn(1:end-1, :));
      if ~isempty(this.u)
        xRes = xRes - this.u(2:end, :);
      end
      xLogProb = stateTransitionNoise.LogPdf(xRes); % xLogProb: T-1 double vector
      logLikelihood = sum(xLogProb);
    end

    function logLikelihood = MHObsLogLikelihood( ...
        this, ...
        fnObs, ...
        observationNoise)

      % Calculate the joint probability of the sequence x(1:T) and y(1:T)
      % under the state space model:
      %   x(t) = fnState([x(t-1) xaux(t-1)]) + u(t) + q
      %   y(t+K) = fnObs(x(t)) + r
      %   q ~ stateTransitionNoise
      %   r ~ observationNoise
      % Notes
      %  1. for q and r, normal distribution is used unless otherwise specified.
      %  2. K is introduced as 'observation offset,' but we use K=0 usually.

      xe = this.GetCurrentXE();

      if nargin < 3
        observationNoise = NormalDistribution(0, this.r, size(this.y, 2));
      end
      xIn      = [xe this.xaux];
      yRes     = this.y(this.K+1:end, :) - fnObs(xIn(1:end-this.K, :));
      yLogProb = observationNoise.LogPdf(yRes); % yLogProb: T double vector
      logLikelihood = sum(yLogProb);
    end

  end

end
