%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% ParticleFilter.m
%

classdef ParticleFilter < LikelihoodCalculator

  properties (Access = protected)
    ancestorIndicies
  end

  properties (SetAccess = protected)
    Particles
    Weights
    Loglik
  end

  properties (GetAccess = public, SetAccess = private)
    x0
    xaux
    u
    y
    q
    r
    N
    K
    OGC   % Out of the Grid Count
  end

  properties (Access = private)
    currentStateTransitionFunction
    currentObservationFunction
    ji
    grids
  end

  methods (Access = public)

    % *** GWP: NOTE - this is for an SIR filter - i.e sample for each particle
    % at time t for index i from prior for state equation
    % weight by the observation equation (i.e. likelihood)

    % Particle filter
    % Input:
    %   x0 - initial condition (1 * xdim)
    %   xaux - auxiliary condigion (T * adim)
    %   u  - control (T * xdim)
    %   y  - measurements (T * ydim)
    %   q  - process noise variance
    %   r  - measurement noise variance
    %   N  - number of particles
    %   K  - observation offset
    function this = ParticleFilter(x0, xaux, u, y, q, r, N, K)
      this.x0 = x0;
      this.xaux = xaux;
      this.u = u;
      this.y = y;
      this.q = q;
      this.r = r;
      this.N = N;
      this.K = K;
      this.OGC = [];
      this.grids = [];
    end

    function SetDebugGrids(this, grids)
      this.grids = zeros(length(grids), 2);
      for i = 1:length(grids)
        this.grids(i, 1) = min(grids{i});
        this.grids(i, 2) = max(grids{i});
      end
    end

    function logLikelihood = LogLikelihood( ...
        this, ...
        stateTransitionFunction, ...
        observationFunction, ...
        stateTransitionNoise, ...
        observationNoise)
      if nargin < 4
        stateTransitionNoise = NormalDistribution(0, this.q, size(this.x0, 2));
        observationNoise = NormalDistribution(0, this.r, size(this.y, 2));
      end
      this.Estimate( ...
          stateTransitionFunction, ...
          observationFunction, ...
          stateTransitionNoise, ...
          observationNoise);
      logLikelihood = this.Loglik;
    end

    % Particle filter
    % Input:
    %   f  - state transition function (R^xdim -> R^xdim)
    %   h  - observation function (R^xdim -> R^ydim)
    % Notes:
    %   f  - (N, xdim + adim) -> (N, xdim)
    %   h  - (N, xdim + adim) -> (N, ydim)
    function Estimate( ...
        this, ...
        f, ...
        h, ...
        stateTransitionNoise, ...
        observationNoise)
      x0 = this.x0;
      xaux = this.xaux;
      u = this.u;
      y = this.y;
      N = this.N;
      K = this.K;

      this.OGC(end + 1) = 0;

      this.currentStateTransitionFunction = f;
      this.currentObservationFunction = h;

      % Retrieve dimensions from input data
      xdim = size(x0, 2);
      [T ydim nseq] = size(y);

      % Initialization
      this.Particles = zeros(N, T, xdim);
      this.ancestorIndicies = repmat([1:N]', 1, T);
      this.Weights = zeros(N, T);

      if size(x0, 1) == 1
        this.Particles(:, 1, :) = repmat(x0, N, 1);
      else
        this.Particles(:, 1, :) = x0;
      end
      this.Loglik = 0;

      for t = (K + 1):T

        if t ~= K + 1
          ind = ParticleFilter.resampling(this.Weights(:, t - K - 1));
          if length(xaux) > 0
            xpred = f([squeeze(this.Particles(:, t - K - 1, :)) repmat(xaux(t - K - 1, :), N, 1)]);
          else
            xpred = f(squeeze(this.Particles(:, t - K - 1, :)));
          end
          if length(u) > 0
            xpred = xpred + repmat(u(t - K, :), N, 1);
          end
          this.Particles(:, t - K, :) = xpred(ind, :) + stateTransitionNoise.Random(N);
          this.ancestorIndicies(:, t - K) = ind;

          if ~isempty(this.grids)
            xs = squeeze(this.Particles(:, t - K, :));
            this.OGC(end) = this.OGC(end) + sum(max(bsxfun(@le, xs, this.grids(:, 1)') + bsxfun(@ge, xs, this.grids(:, 2)'), [], 2));
          end
        end

        % Compute importance weights
        if length(xaux) > 0
          ypred = h([squeeze(this.Particles(:, t - K, :)) repmat(xaux(t - K, :), N, 1)]);
        else
          ypred = h(squeeze(this.Particles(:, t - K, :)));
        end

        logweights = zeros(N, 1);
        for i = 1:nseq
          yResidual = repmat(y(t, :, i), N, 1) - ypred;
          logweights = logweights + observationNoise.LogPdf(yResidual);
        end
        const = max(logweights);
        weights = exp(logweights - const);

        % Compute loglikelihood
        this.Loglik = this.Loglik + const + log(sum(weights)) - log(N);
        this.Weights(:, t - K) = weights / sum(weights);

      end
    end

    function ypred = EstimateNext( ...
        this, ...
        xaux, ...
        u, ...
        stateTransitionNoise, ...
        observationNoise)

      if nargin < 4
        stateTransitionNoise = NormalDistribution(0, this.q, size(this.x0, 2));
        observationNoise = NormalDistribution(0, this.r, size(this.y, 2));
      end

      x0 = this.x0;
      xaux = [this.xaux; xaux];
      u = [this.u; u];
      q = this.q;
      r = this.r;
      N = this.N;
      K = this.K;

      f = this.currentStateTransitionFunction;
      h = this.currentObservationFunction;

      T = size(this.Particles, 2);

      ind = ParticleFilter.resampling(this.Weights(:, T - K));
      if length(xaux) > 0
        xpred = f([squeeze(this.Particles(:, T - K, :)) repmat(xaux(T - K, :), N, 1)]);
      else
        xpred = f(squeeze(this.Particles(:, T - K, :)));
      end
      if length(u) > 0
        xpred = xpred + repmat(u(T - K + 1, :), N, 1);
      end
      this.Particles(:, T - K + 1, :) = xpred(ind, :) + stateTransitionNoise.Random(N);

      if length(xaux) > 0
        ypred = h([squeeze(this.Particles(:, T - K + 1, :)) repmat(xaux(T - K + 1, :), N, 1)]);
      else
        ypred = h(squeeze(this.Particles(:, T - K + 1, :)));
      end
      ypred = ypred + observationNoise.Random(N);

      this.Weights(:, T - K + 1) = 1;

    end

    function OnAccept(this, j, in, out, aspect)
      ji = find(rand(1) < cumsum(this.Weights(:, in.T() - in.K)), 1, 'first');
      xe = reshape(this.Particles(ji, :, :), size(this.Particles, 2), size(this.Particles, 3));
      out.AddXE(xe);
      if length(in.xaux) > 0
        out.AddYE(this.currentObservationFunction([xe in.xaux']));
      else
        out.AddYE(this.currentObservationFunction(xe));
      end
      out.AddLoglik(this.Loglik);
      out.AddWeight(this.Weights);
      this.ji = ji;
    end

  end

  methods (Access = protected)

    function xe = GetCurrentXE(this)
      xe = reshape(this.Particles(this.ji, :, :), size(this.Particles, 2), size(this.Particles, 3));
    end

  end

  methods (Static, Access = private)

    % ======================================================================
    %
    %   Multinomial resampling function
    %
    % ======================================================================
    function i = resampling(q)
      N = length(q);
      % Multinomial resampling
      u = rand(N, 1);
      qc = cumsum(q);
      qc = qc(:);
      qc = qc / qc(N);
      [~, ind1] = sort([u; qc]);
      ind2 = find(ind1 <= N);
      i = ind2' - (0:N-1);
    end

  end

end
