%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% GeneralizedParetoDistributionTest.m
%
classdef GeneralizedParetoDistributionTest < matlab.unittest.TestCase

  methods (Test)

    function testGeneralizedParetoDistribution(testCase)
      dist = GeneralizedParetoDistribution(1.5, 2.5, 3.5);
      testCase.verifyEqual(dist.K, 1.5);
      testCase.verifyEqual(dist.Sigma, 2.5);
      testCase.verifyEqual(dist.Theta, 3.5);
    end

    function testPdfKZero(testCase)
      dist = GeneralizedParetoDistribution(0, 2.5, 3.5);
      actual = dist.Pdf(4.5);
      expected = gppdf(4.5, 0, 2.5, 3.5);
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testPdfKPos(testCase)
      dist = GeneralizedParetoDistribution(1.5, 2.5, 3.5);
      actual = dist.Pdf(4.5);
      expected = gppdf(4.5, 1.5, 2.5, 3.5);
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testPdfKPosLB(testCase)
      dist = GeneralizedParetoDistribution(1.5, 2.5, 3.5);
      actual = dist.Pdf(3.35);
      expected = 0;
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testPdfKNeg(testCase)
      dist = GeneralizedParetoDistribution(-1.5, 2.5, 3.5);
      actual = dist.Pdf(4.5);
      expected = gppdf(4.5, -1.5, 2.5, 3.5);
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testPdfKNegLB(testCase)
      dist = GeneralizedParetoDistribution(-1.5, 2.5, 3.5);
      actual = dist.Pdf(-0.3);
      expected = 0;
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testPdfKNegUB(testCase)
      dist = GeneralizedParetoDistribution(-1.5, 2.5, 3.5);
      actual = dist.Pdf(5.5);
      expected = 0;
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testPdfMultiVariate(testCase)
      dist = GeneralizedParetoDistribution(1.5, 2.5, 3.5, 3);
      x = [1.35 2.35 3.35; 3.5 3.6 3.7; 4 4 4];
      actual = dist.Pdf(x);
      expected = [0; prod(gppdf(x(2:3, :), 1.5, 2.5, 3.5), 2)];
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testLogPdfKZero(testCase)
      dist = GeneralizedParetoDistribution(0, 2.5, 3.5);
      actual = dist.LogPdf(4.5);
      expected = log(gppdf(4.5, 0, 2.5, 3.5));
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testLogPdfKPos(testCase)
      dist = GeneralizedParetoDistribution(1.5, 2.5, 3.5);
      actual = dist.LogPdf(4.5);
      expected = log(gppdf(4.5, 1.5, 2.5, 3.5));
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testLogPdfKPosLB(testCase)
      dist = GeneralizedParetoDistribution(1.5, 2.5, 3.5);
      actual = dist.LogPdf(3.35);
      expected = -Inf;
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testLogPdfKNeg(testCase)
      dist = GeneralizedParetoDistribution(-1.5, 2.5, 3.5);
      actual = dist.LogPdf(4.5);
      expected = log(gppdf(4.5, -1.5, 2.5, 3.5));
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testLogPdfKNegLB(testCase)
      dist = GeneralizedParetoDistribution(-1.5, 2.5, 3.5);
      actual = dist.LogPdf(-0.3);
      expected = -Inf;
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testLogPdfKNegUB(testCase)
      dist = GeneralizedParetoDistribution(-1.5, 2.5, 3.5);
      actual = dist.LogPdf(5.5);
      expected = -Inf;
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testLogPdfMultiVariate(testCase)
      dist = GeneralizedParetoDistribution(1.5, 2.5, 3.5, 3);
      x = [1.35 2.35 3.35; 3.5 3.6 3.7; 4 4 4];
      actual = dist.LogPdf(x);
      expected = [-Inf; log(prod(gppdf(x(2:3, :), 1.5, 2.5, 3.5), 2))];
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

  end

end
