%%% Monte Carlo Dynamic Classifier tool - MCDC tool

% Copyright (C) 2014 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
% Floor, Boston, MA 02110-1301 USA.

%
% LaplaceDistributionTest.m
%
classdef LaplaceDistributionTest < matlab.unittest.TestCase

  methods (Test)

    function testLaplaceDistribution(testCase)
      dist = LaplaceDistribution(2, 3);
      testCase.verifyEqual(dist.Location, 2);
      testCase.verifyEqual(dist.Scale, 3);
    end

    function testPdf(testCase)
      dist = LaplaceDistribution(2, 3);
      actual = dist.Pdf(2);
      expected = 1 / (2 * 3);
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testPdfMeanPlus2Sigma(testCase)
      dist = LaplaceDistribution(2, 3);
      actual = dist.Pdf(8);
      expected = (1/6) * exp(-2);
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testPdfMeanMinus3Sigma(testCase)
      dist = LaplaceDistribution(2, 3);
      actual = dist.Pdf(-7);
      expected = (1/6) * exp(-3);
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testPdfMultiVariate(testCase)
      dist = LaplaceDistribution(2, 3, 3);
      x = [0 0 0; 1 2 3; 5 4 3];
      actual = dist.Pdf(x);
      edist = LaplaceDistribution(2, 3);
      expected = [ ...
        edist.Pdf(0) * edist.Pdf(0) * edist.Pdf(0); ...
        edist.Pdf(1) * edist.Pdf(2) * edist.Pdf(3); ...
        edist.Pdf(5) * edist.Pdf(4) * edist.Pdf(3)  ...
      ];
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testLogPdf(testCase)
      dist = LaplaceDistribution(2, 3);
      actual = dist.LogPdf(0);
      expected = log(dist.Pdf(0));
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testLogPdfMultiVariate(testCase)
      dist = LaplaceDistribution(2, 3, 3);
      x = [0 0 0; 1 2 3; 5 4 3];
      actual = dist.LogPdf(x);
      expected = log(dist.Pdf(x));
      testCase.verifyEqual(actual, expected, 'RelTol', 1e-6);
    end

    function testRandom(testCase)
      dist = LaplaceDistribution(2, 3);
      x = dist.Random(1000000);
      % TODO: how large tolerance we should allow?
      testCase.verifyEqual(mean(x), 2, 'AbsTol', 1e-1);
      testCase.verifyEqual(var(x), 2 * 3 ^ 2, 'AbsTol', 1e-1);
      testCase.verifyEqual(skewness(x), 0, 'AbsTol', 1e-1);
      testCase.verifyEqual(kurtosis(x), 6, 'AbsTol', 1e-1);
    end

    function testRandomMultiVariate(testCase)
      dist = LaplaceDistribution(2, 3, 3);
      x = dist.Random(1000000);
      % TODO: how large tolerance we should allow?
      testCase.verifyEqual(mean(x), [2 2 2], 'AbsTol', 1e-1);
      testCase.verifyEqual(var(x), [18 18 18], 'AbsTol', 1e-1);
      testCase.verifyEqual(skewness(x), [0 0 0], 'AbsTol', 1e-1);
      testCase.verifyEqual(kurtosis(x), [6 6 6], 'AbsTol', 1e-1);
    end

  end

end
