%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% KitagawaModelPMCMC2Estimation.m
%


% preconditions
rng('default');
warning off

%
% Generate state and observation time series from Kitagawa model
%

T  = 10;
a1 = 0.5;
b1 = 28;
c1 = 8;
a2 = 0.6;
b2 = 30;
c2 = 10;
d1 = 0.05;
d2 = 0.06;
d3 = 0.07;
d4 = 0.08;
q  = 1e-1;
r  = 1e-1;

[x, y] = KitagawaModel(T, a1, b1, c1, a2, b2, c2, d1, d2, d3, d4, q, r);

save('logs/KitagawaModelPMCMC2_Data_UNIT_TEST_RUN.mat');

grids = { ...
  [-30:2:30], ...
  [-30:2:30]  ...
};

stateKernelGens = { ...
  RBFKernelGenerator(UniformDistribution(0.01, 10), UniformDistribution(0.01, 10)), ...
  RBFKernelGenerator(UniformDistribution(0.01, 10), UniformDistribution(0.01, 10))  ...
};

obsKernelGens = {
  RBFKernelGenerator(UniformDistribution(0.01, 10), UniformDistribution(0.01, 10)), ...
  RBFKernelGenerator(UniformDistribution(0.01, 10), UniformDistribution(0.01, 10)), ...
  RBFKernelGenerator(UniformDistribution(0.01, 10), UniformDistribution(0.01, 10))  ...
};

stateMeanFuncs = { ...
  @(x) (a1 .* x(1,:) + b1 .* x(1,:) / (1 + x(1,:) .^ 2)), ...
  @(x) (a2 .* x(2,:) + b2 .* x(2,:) / (1 + x(2,:) .^ 2))  ...
};

obsMeanFuncs = { ...
  @(x) (d1 .* x(1,:) .^ 2 + d2 .* x(2,:) .^ 2), ...
  @(x) (d3 .* x(1,:) .^ 2                    ), ...
  @(x) (                    d4 .* x(2,:) .^ 2)  ...
};

J  = 10;

gridDimForGramMatrix = 1; % Use x1 for building gram matrix.

aspect = Aspect();
aspect.LogFileName = ['logs/KitagawaModelPMCMC2' '_UNIT_TEST_RUN.log'];
aspect.MatFileNamePrefix = ['logs/KitagawaModelPMCMC2' '_UNIT_TEST_RUN'];
aspect.SavesIntermediateMat = true;
aspect.IntermediateMatInterval = 1000;

disp('PMCMC2 Estimation using SPAPI');

%%% Choice 1
% kernelGeneratorStrategy = @RBFKernelGeneratorStrategyChoice1;

%%% Choice 2
kernelGeneratorStrategy = @RBFKernelGeneratorStrategyChoice2;

%%% Choice 1
% mcdcSrategy = @MCDCStrategyChoice1;

%%% Choice 2
mcdcStrategy = @MCDCStrategyChoice2;

%%% Choice 3: build gram matrices here.
% for i = 1:size(stateMeanFuncs, 2)
%   stateSigma{i} = eye(length(grids{gridDimForGramMatrix}));
% end
% for i = 1:size(obsMeanFuncs, 2)
%   obsSigma{i} = eye(length(grids{gridDimForGramMatrix}));
% end
% mcdcStrategy = MCDCStrategyChoice3.Prepare(stateSigma, obsSigma);

stateSplineHandle = @GenericSpline;
obsSplineHandle = @GenericSpline;
algorithm = PMCMC2(kernelGeneratorStrategy, mcdcStrategy);
algorithm.SetPMCMCProbability(1.0); % Don't use MH-Gibbs

stateModel = GaussianProcessModel(ModelKind.State, algorithm);
algorithm.SetStateModel(stateModel);

obsModel = GaussianProcessModel(ModelKind.Observation, algorithm);
algorithm.SetObsModel(obsModel);

x0 = zeros(1, 2);
xaux = [];
u = repmat(cos(1.2 * [1:T]), 2, 1)';
q = 1;
r = 0.1;
N = 10;
likelihoodCalculator = PMCMCParticleFilter(x0, xaux, u, y, q, r, N, 0);
algorithm.SetLikelihoodCalculator(likelihoodCalculator);


%% Test 1: set up model correctly
assert(T > 0, 'Problem with number Time steps should be positive integer')
assert(N > 0, 'Problem with number particles should be positive integer')
assert(isempty(a1) == 0, 'Problem with parameter a1 should have positive real value') 
assert(isempty(b1) == 0, 'Problem with parameter b1 should have positive real value') 
assert(isempty(c1) == 0, 'Problem with parameter c1 should have positive real value') 
assert(isempty(a2) == 0, 'Problem with parameter a2 should have positive real value') 
assert(isempty(b2) == 0, 'Problem with parameter b2 should have positive real value') 
assert(isempty(c2) == 0, 'Problem with parameter c2 should have positive real value') 
assert(isempty(d1) == 0, 'Problem with parameter d1 should have positive real value') 
assert(isempty(d2) == 0, 'Problem with parameter d2 should have positive real value') 
assert(isempty(d3) == 0, 'Problem with parameter d3 should have positive real value') 
assert(isempty(d4) == 0, 'Problem with parameter d4 should have positive real value') 
assert(isempty(q) == 0, 'Problem with parameter q should have positive real value') 
assert(isempty(r) == 0, 'Problem with parameter r should have positive real value') 
c = size(x);
assert(length(x) == T, 'State process data not generated correctly')
assert(c(2) == 2, 'dimension of state vector incorrect')
assert(length(y) == T, 'Observation process data not generated correctly')
c = size(y);
assert(c(2) == 3, 'dimension of observation vector incorrect')


%% Test 2: set up kernel specifications for Gaussian Process State

val1 = stateKernelGens{1,1}.SDistribution.LowerBound;
val2 = stateKernelGens{1,1}.SDistribution.UpperBound;
val3 = stateKernelGens{1,1}.LDistribution.LowerBound;
val4 = stateKernelGens{1,1}.LDistribution.UpperBound;
assert(val1 > 0, 'Lower Bound State GP distribution Kernel Param 1 should be positive')
assert(isinf(val2) == 0 , 'Upper Bound State GP distribution Kernel Param 1 should be finite')
assert(val3 > 0, 'Lower Bound State GP distribution Kernel Param 2 should be positive')
assert(isinf(val4) == 0 , 'Upper Bound State GP distribution Kernel Param 2 should be finite')

%% Test 3: set up kernel specifications for Gaussian Process Observations

val1 = stateKernelGens{1,1}.SDistribution.LowerBound;
val2 = stateKernelGens{1,1}.SDistribution.UpperBound;
val3 = stateKernelGens{1,1}.LDistribution.LowerBound;
val4 = stateKernelGens{1,1}.LDistribution.UpperBound;
assert(val1 > 0, 'Lower Bound Observation GP distribution Kernel Param 1 should be positive')
assert(isinf(val2) == 0 , 'Upper Bound Observation GP distribution Kernel Param 1 should be finite')
assert(val3 > 0, 'Lower Bound Observation GP distribution Kernel Param 2 should be positive')
assert(isinf(val4) == 0 , 'Upper Bound Observation GP distribution Kernel Param 2 should be finite')

%% Test 4: test functions in PMCMC2 Code

[ IDX, SKP, OKP, FV, GV, XE, YE, loglik ] = MCDCTrain( ...
  algorithm, ...
  grids, ...
  stateKernelGens, ...
  obsKernelGens, ...
  stateMeanFuncs, ...
  obsMeanFuncs, ...
  gridDimForGramMatrix, ...
  stateSplineHandle, ...
  obsSplineHandle, ...
  x0, ...
  [], ... % No auxiliary states
  u', ...
  y', ...
  N, ...
  J, ...
  0, ...
  aspect ...
);

assert(length(XE(:,1,1)) == 1, 'Incorrect dimension of state vectors created')
assert(length(XE(1,:,1)) == T, 'Incorrect length of state vectors created')
assert(length(YE(:,1,1)) == 1, 'Incorrect number of particles utilised')
assert(length(YE(1,:,1)) == T, 'Incorrect length of state vectors created')
assert(isnan(sum(loglik))== 0,'Problem with log Likelihood calculation')

