%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% LinearStateSpaceModelPSMCEstimation.m
%

% preconditions
rng('default');
warning off

%
% Generate state and observation time series from Linear state space model
%

T    = 10;
k1   = 0.1;
k2   = 0.1;
mass = 1.0;
q    = 0.1;
r    = 0.1;
dt   = 1.0;

[x, y, F, H] = LinearStateSpaceModel(T, k1, k2, mass, q, r, dt);

save('logs/LinearStasteSpaceModel_Data.mat');

grids = { ...
  [-2:0.1:2], ...
  [-2:0.1:2]  ...
};

stateKernelGens = { ...
  RBFKernelGenerator(UniformDistribution(0.01, 10), UniformDistribution(0.01, 10)), ...
  RBFKernelGenerator(UniformDistribution(0.01, 10), UniformDistribution(0.01, 10))  ...
};

obsKernelGens = {
  RBFKernelGenerator(UniformDistribution(0.01, 10), UniformDistribution(0.01, 10))  ...
};

stateMeanFuncs = { ...
  @(x) x(1,:), ...
  @(x) x(2,:)  ...
};

obsMeanFuncs = { ...
  @(x) x(1,:)  ...
};

x0 = zeros(1, 2);
xaux = []; % No auxiliary states
u = [];    % No controls

N  = 10;
J  = 10;

aspect = Aspect();
aspect.LogFileName = 'logs/LinearStateSpaceModelPSMC.log';
aspect.MatFileNamePrefix = 'logs/LinearStateSpaceModelPSMC';
aspect.SavesIntermediateMat = true;

disp('PSMC Estimation using SPAPI');

stateSplineHandle = @GenericSpline;
obsSplineHandle = @GenericSpline;
algorithm = PSMC();
algorithm.SetPMCMCProbability(1.0); % Don't use MH-Gibbs

stateModel = GaussianProcessModel(ModelKind.State, algorithm);
algorithm.SetStateModel(stateModel);

obsModel = GaussianProcessModel(ModelKind.Observation, algorithm);
algorithm.SetObsModel(obsModel);

likelihoodCalculator = ParticleFilter(x0, xaux, u, y, q, r, N, 0);
algorithm.SetLikelihoodCalculator(likelihoodCalculator);


%% Test 1: set up model correctly
assert(T > 0, 'Problem with number Time steps should be positive integer')
assert(N > 0, 'Problem with number particles should be positive integer')
assert(k1 > 0, 'k1 parameter should be positive')
assert(k2 > 0, 'k1 parameter should be positive')
c = size(x);
assert(length(x) == T, 'State process data not generated correctly')
assert(c(2) == 2, 'dimension of state vector incorrect')
assert(length(y) == T, 'Observation process data not generated correctly')
c = size(y);
assert(c(2) == 1, 'dimension of observation vector incorrect')


%% Test 2: set up kernel specifications for Gaussian Process State

val1 = stateKernelGens{1,1}.SDistribution.LowerBound
val2 = stateKernelGens{1,1}.SDistribution.UpperBound
val3 = stateKernelGens{1,1}.LDistribution.LowerBound
val4 = stateKernelGens{1,1}.LDistribution.UpperBound
assert(val1 > 0, 'Lower Bound State GP distribution Kernel Param 1 should be positive')
assert(isinf(val2) == 0 , 'Upper Bound State GP distribution Kernel Param 1 should be finite')
assert(val3 > 0, 'Lower Bound State GP distribution Kernel Param 2 should be positive')
assert(isinf(val4) == 0 , 'Upper Bound State GP distribution Kernel Param 2 should be finite')

%% Test 3: set up kernel specifications for Gaussian Process Observations

val1 = stateKernelGens{1,1}.SDistribution.LowerBound
val2 = stateKernelGens{1,1}.SDistribution.UpperBound
val3 = stateKernelGens{1,1}.LDistribution.LowerBound
val4 = stateKernelGens{1,1}.LDistribution.UpperBound
assert(val1 > 0, 'Lower Bound Observation GP distribution Kernel Param 1 should be positive')
assert(isinf(val2) == 0 , 'Upper Bound Observation GP distribution Kernel Param 1 should be finite')
assert(val3 > 0, 'Lower Bound Observation GP distribution Kernel Param 2 should be positive')
assert(isinf(val4) == 0 , 'Upper Bound Observation GP distribution Kernel Param 2 should be finite')

%% Test 4: test functions in PSMC Linear Kalman Filter Code

[ IDX, SKP, OKP, FV, GV, XE, YE, loglik ] = MCDCTrain( ...
  algorithm, ...
  grids, ...
  stateKernelGens, ...
  obsKernelGens, ...
  stateMeanFuncs, ...
  obsMeanFuncs, ...
  0, ... % Not used in PSMC estimation
  stateSplineHandle, ...
  obsSplineHandle, ...
  x0, ...
  [], ... % No auxiliary states
  [], ... % No controls
  y', ...
  N, ...
  J, ...
  0, ...
  aspect ...
);

assert(length(XE(:,1,1)) == N, 'Incorrect dimension of state vectors created')
assert(length(XE(1,:,1)) == T, 'Incorrect length of state vectors created')
assert(length(YE(:,1,1)) == N, 'Incorrect number of particles utilised')
assert(length(YE(1,:,1)) == T, 'Incorrect length of state vectors created')
assert(isnan(sum(loglik))== 0,'Problem with log Likelihood calculation')



